#!/usr/bin/env python

"""
View annotated sources.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from compiler.visitor import ASTVisitor
from simplified import *
import sys
import os
import textwrap

# Exceptions.

class ViewerError(SimplifiedError):

    "An error in viewing."

    pass

# Classes.

class Viewer(ASTVisitor):

    """
    A viewing visitor for AST nodes.
    """

    def __init__(self, stream):
        ASTVisitor.__init__(self)
        self.cached_files = {}
        self.printed_lines = {}
        self.visitor = self
        self.stream = stream

    def process(self, module):
        self.dispatch(module)

    def dispatch(self, node):
        self.dispatch_only(node)
        ASTVisitor.dispatch(self, node)

    def dispatch_only(self, node, every_time=0):
        self.print_line(getattr(node, "filename", None), getattr(node, "lineno", None), every_time)

    def print_line(self, filename, lineno, every_time):
        last_printed = self.printed_lines.get(filename, 0)
        if lineno > last_printed or every_time:
            self.stream.write(self.get_line(filename, lineno))
            self.printed_lines[filename] = lineno

    def get_line(self, filename, lineno):
        if filename is None or lineno is None:
            return ""

        if self.cached_files.has_key(filename):
            lines = self.cached_files[filename]
        else:
            f = open(filename)
            try:
                self.cached_files[filename] = lines = f.readlines()
            finally:
                f.close()

        try:
            return lines[lineno - 1]
        except IndexError:
            return ""

    def report(self, exc):
        self.stream.write("Exception was:\n\n" + str(exc.exc) + "\n\n")
        self.stream.write("Nodes:\n\n")
        for node in exc.nodes:
            self.stream.write(repr(node) + "\n")
            if node is not None and hasattr(node, "original"):
                self.dispatch_only(node.original, every_time=1)
        if hasattr(exc.nodes[0], "original"):
            self.stream.write("\nOriginal node was:\n\n" + repr(exc.nodes[0].original) + "\n")
        self.stream.write("\nSimplified node was:\n\n")
        exc.nodes[0].pprint(stream=self.stream)

# HTML-related output production.

html_header = """<?xml version="1.0" encoding="iso-8859-15"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
  <title>Module</title>
  <style type="text/css">
    body {
      padding-top: 4em; padding-bottom: 4em;
      font-size: 14pt; font-family: monospace;
      background-color: black; color: white;
    }

    .class { margin-bottom: 1em; }
    .function { margin-bottom: 1em; }
    .body { padding-left: 2em; }
    .keyword { color: yellow; }
    .comment { color: blue; }
    .str { color: #FF00FF; }
    .doc { color: #FF00FF; margin-bottom: 1em; }
    .invocation a { color: white; text-decoration: none; }

    .popup {
      display: none; z-index: 2;
      position: absolute; top: 1em; left: 0.5em;
      padding: 0.2em; background-color: #000000;
      border: 2px solid #dddddd;
    }

    .invocations {
      padding: 0.5em; background-color: #770000;
      clear: all;
    }

    .types {
      padding: 0.5em; background-color: #0000FF;
      float: right;
    }

    .raises {
      padding: 0.5em; background-color: #7700FF;
      float: right;
    }

    .scopes {
      padding: 0.5em; background-color: #007700;
      float: left;
    }

    .non-writes, .non-accesses {
      padding: 0.5em; background-color: #FF0000;
      float: right;
    }

    .op,
    .name,
    .attr,
    .conditional,
    .operator,
    .iterator,
    .returns
    {
      position: relative;
    }

    .op:hover > .popup,
    .name:hover > .popup,
    .attr:hover > .popup,
    .conditional:hover > .popup,
    .operator:hover > .popup,
    .iterator:hover > .popup,
    .returns:hover > .popup
    {
       display: block;
    }

  </style>
</head>
<body>
"""

html_footer = """</body>
</html>
"""

# Browser classes.

class Browser(ASTVisitor):

    """
    A browsing visitor for AST nodes.

    Covered: Add, And, AssAttr, AssList, AssName, AssTuple, Assign, AugAssign,
             Break, CallFunc, Class, Compare, Const, Continue, Dict, Discard,
             Div, FloorDiv, For, Function, Getattr, If, Keyword, Lambda, List,
             Module, Mul, Name, Not, Or, Pass, Raise, Return, Slice, Stmt, Sub,
             Subscript, TryExcept, TryFinally, Tuple, UnaryAdd, UnarySub, While.

    Missing: Assert, Backquote, Bitand, Bitor, Bitxor, Decorators, Ellipsis,
             Exec, From, Global, Import, Invert, LeftShift, ListComp,
             ListCompFor, ListCompIf, Mod, Power, Print, Printnl, RightShift,
             Sliceobj, Yield.
    """

    def __init__(self, stream):
        ASTVisitor.__init__(self)
        self.visitor = self
        self.stream = stream

    def process(self, module):
        self.stream.write(html_header)
        self.dispatch(module)
        self.stream.write(html_footer)

    def dispatch(self, node):
        try:
            ASTVisitor.dispatch(self, node)
        except ViewerError, exc:
            exc.add(node)
            raise
        except Exception, exc:
            raise ViewerError(exc, node)

    def visitModule(self, node):
        self.default(node)

    # Statements.

    def visitAssign(self, node):
        self.stream.write("<div class='assign'>\n")
        for lvalue in node.nodes:
            self.dispatch(lvalue)
            self.stream.write("=\n")
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitAugAssign(self, node):
        self.stream.write("<div class='augassign'>\n")
        self.dispatch(node.node)
        self.stream.write("%s\n" % node.op)
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitBreak(self, node):
        self.stream.write("<div class='break'>\n")
        self._keyword("break")
        self.stream.write("</div>\n")

    def visitClass(self, node):
        definition = node._node
        structure = definition.expr.ref
        self.stream.write("<div class='class' id='%s'>\n" % self._url(structure.full_name()))
        self.stream.write("<div>\n")
        self._keyword("class")
        self._name_start(structure.name)
        self._popup_start()
        self._scopes(definition)
        self._popup_end()
        self._name_end()
        bases = structure.bases

        # Suppress the "object" class appearing alone.

        if bases and not (len(bases) == 1 and bases[0].name == "object"):
            self.stream.write("(")
            first = 1
            for base in bases:
                if not first:
                    self.stream.write(",\n")
                self._name_start(base.name)
                self._popup_start()
                self._types(base)
                self._scopes(base)
                self._popup_end()
                self._name_end()
                first = 0
            self.stream.write(")")

        self.stream.write(":\n")
        self._comment(self._text(structure.full_name()))
        self.stream.write("</div>\n")

        self.stream.write("<div class='body'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitContinue(self, node):
        self.stream.write("<div class='continue'>\n")
        self._keyword("continue")
        self.stream.write("</div>\n")

    def visitDiscard(self, node):
        self.stream.write("<div class='discard'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitFor(self, node):
        self.stream.write("<div class='if'>\n")
        self.stream.write("<div>\n")
        self.stream.write("<span class='iterator'>\n")
        self._keyword("for")
        self._popup_start()
        self._invocations(node._node.code[1].expr.ref.code[0].body[0].code[0].expr) # Link to next call in subprogram.
        self._popup_end()
        self.stream.write("</span>\n")
        self.dispatch(node.assign)
        self.stream.write("<span class='iterator'>\n")
        self._keyword("in")
        self._popup_start()
        self._invocations(node._node.code[0].expr) # Link to __iter__ call.
        self._popup_end()
        self.stream.write("</span>\n")
        self.dispatch(node.list)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitFunction(self, node):
        definition = node._node
        subprogram = definition.expr.ref
        self.stream.write("<div class='function' id='%s'>\n" % self._url(subprogram.full_name()))
        self.stream.write("<div>\n")
        self._keyword("def")
        self._name_start(subprogram.name)
        self._popup_start()
        self._scopes(definition)
        self._raises(subprogram)
        self._popup_end()
        self._name_end()
        self.stream.write("(")
        self._parameters(subprogram)
        self.stream.write(")")
        self.stream.write(":\n")
        self._comment(self._text(subprogram.full_name()))
        self.stream.write("</div>\n")

        self.stream.write("<div class='body'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitIf(self, node):
        self.stream.write("<div class='if'>\n")
        first = 1
        conditional = node._node
        for compare, stmt in node.tests:
            self.stream.write("<div>\n")
            self.stream.write("<span class='conditional'>\n")
            if first:
                self._keyword("if")
            else:
                self._keyword("elif")
            self._popup_start()
            self._invocations(conditional.test)
            self._popup_end()
            self.stream.write("</span>\n")
            self.dispatch(compare)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(stmt)
            self.stream.write("</div>\n")
            if conditional.else_:
                conditional = conditional.else_[0]
            else:
                conditional = None
            first = 0
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitPass(self, node):
        self.stream.write("<div class='pass'>\n")
        self._keyword("pass")
        self.stream.write("</div>\n")

    def visitRaise(self, node):
        self.stream.write("<div class='raise'>\n")
        self._keyword("raise")
        self.dispatch(node.expr1)
        if node.expr2 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr2)
        if node.expr3 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr3)
        self.stream.write("</div>\n")

    def visitReturn(self, node):
        self.stream.write("<div class='return'>\n")
        self.stream.write("<span class='returns'>\n")
        self._keyword("return")
        self._popup_start()
        self._types(node._node)
        self._popup_end()
        self.stream.write("</span>\n")
        self.dispatch(node.value)
        self.stream.write("</div>\n")

    def visitStmt(self, node):
        self.stream.write("<div class='stmt'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitTryExcept(self, node):
        self.stream.write("<div class='tryexcept'>\n")
        self.stream.write("<div>\n")
        self._keyword("try")
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        for spec, assign, statement in node.handlers:
            self.stream.write("<div>\n")
            self._keyword("except")
            if spec is not None:
                self.dispatch(spec)
            if assign is not None:
                self.stream.write(",\n")
                self.dispatch(assign)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(statement)
            self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitTryFinally(self, node):
        self.stream.write("<div class='tryfinally'>\n")
        self.stream.write("<div>\n")
        self._keyword("try")
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        self.stream.write("<div>\n")
        self._keyword("finally")
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.final)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitWhile(self, node):
        self.stream.write("<div class='while'>\n")
        self.stream.write("<div>\n")
        self.stream.write("<span class='conditional'>\n")
        self._keyword("while")
        self._popup_start()
        self._invocations(node.test)
        self._popup_end()
        self.stream.write("</span>\n")
        self.dispatch(node.test)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    # Expression-related helper methods.

    def _visitBinary(self, node, name, symbol):
        self.stream.write("<span class='%s'>\n" % name)
        self.dispatch(node.left)
        self.stream.write("<span class='operator'>\n")
        self.stream.write(symbol)
        self._popup_start()
        self.stream.write("<div class='invocations'>\n")
        self._invocations_list(node._node.body[0].expr) # NOTE: See visitAdd in simplify.
        self._invocations_list(node._node.handler[0].body[0].expr) # NOTE: See visitAdd in simplify.
        self.stream.write("</div>\n")
        self._popup_end()
        self.stream.write("</span>\n")
        self.dispatch(node.right)
        self.stream.write("</span>")

    def _visitUnary(self, node, name, symbol):
        self.stream.write("<span class='%s'>\n" % name)
        self.stream.write("<span class='operator'>\n")
        self.stream.write(symbol)
        self._popup_start()
        self.stream.write("<div class='invocations'>\n")
        self._invocations_list(node._node) # NOTE: See _visitUnary in simplify.
        self.stream.write("</div>\n")
        self._popup_end()
        self.stream.write("</span>\n")
        self.dispatch(node.expr)
        self.stream.write("</span>")

    # Expressions.

    def visitAdd(self, node):
        self._visitBinary(node, "add", "+")

    def visitAnd(self, node):
        self.stream.write("<span class='and'>\n")
        first = 1
        for n in node.nodes:
            if not first:
                self._keyword("and")
            self.dispatch(n)
            first = 0
        self.stream.write("</span>")

    def visitAssAttr(self, node):
        self.stream.write("<span class='assattr'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>\n")
        self.stream.write(".%s\n" % self._text(node.attrname))
        if hasattr(node, "_node"):
            self._popup_start()
            self._types(node._node)
            self._scopes(node._node)
            self._popup_end()
        else:
            raise ValueError, node
        self.stream.write("</span>\n")
        self.stream.write("</span>\n")

    def visitAssList(self, node):
        self.stream.write("<span class='list'>\n")
        self.stream.write("[")
        self._sequence(node)
        self.stream.write("]\n")
        self.stream.write("</span>\n")

    def visitAssName(self, node):
        if hasattr(node, "_node"):
            self._name_start(node._node.name)
            self._popup_start()
            self._types(node._node.expr)
            self._scopes(node._node)
            self._popup_end()
            self._name_end()
        else:
            raise ValueError, node
            self._name(node.name)

    def visitAssTuple(self, node):
        self.stream.write("<span class='tuple'>\n")
        self.stream.write("(")
        self._sequence(node)
        self.stream.write(")\n")
        self.stream.write("</span>\n")

    def visitCallFunc(self, node):
        self.stream.write("<span class='callfunc'>\n")
        self.dispatch(node.node)
        self.stream.write("(")
        first = 1
        for arg in node.args:
            if not first:
                self.stream.write(",\n")
            self.dispatch(arg)
            first = 0
        if node.star_args is not None:
            if not first:
                self.stream.write(", *\n")
            self.dispatch(node.star_args)
            first = 0
        if node.dstar_args is not None:
            if not first:
                self.stream.write(", **\n")
            self.dispatch(node.dstar_args)
            first = 0
        self.stream.write(")\n")
        self.stream.write("</span>\n")

    def visitCompare(self, node):
        self.stream.write("<span class='compare'>\n")
        self.dispatch(node.expr)
        for (op_name, expr), _op in map(None, node.ops, node._ops):
            self.stream.write("<span class='op'>\n")
            self.stream.write(op_name)
            self._popup_start()
            self._op(op_name, _op)
            self._popup_end()
            self.stream.write("</span>\n")
            self.dispatch(expr)
        self.stream.write("</span>\n")

    def visitConst(self, node):
        self.stream.write(repr(node.value))

    def visitDiv(self, node):
        self._visitBinary(node, "div", "/")

    def visitFloorDiv(self, node):
        self._visitBinary(node, "floordiv", "//")

    def visitGetattr(self, node):
        self.stream.write("<span class='getattr'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>\n")
        self.stream.write(".%s\n" % self._text(node.attrname))
        if hasattr(node, "_node"):
            self._popup_start()
            self._types(node._node)
            self._scopes(node._node)
            self._popup_end()
        else:
            raise ValueError, node
        self.stream.write("</span>\n")
        self.stream.write("</span>\n")

    def visitKeyword(self, node):
        self.stream.write("<span class='keyword'>\n")
        self.stream.write(node.name)
        self.stream.write("=")
        self.dispatch(node.expr)
        self.stream.write("</span>\n")

    def visitLambda(self, node):
        definition = node._node
        subprogram = definition.expr.ref
        self.stream.write("<span class='lambda'>\n")
        self._keyword("lambda")
        self._parameters(subprogram)
        self.dispatch(node.code)
        self.stream.write("</span>\n")

    visitList = visitAssList

    def visitMul(self, node):
        self._visitBinary(node, "mul", "*")

    def visitName(self, node):
        if hasattr(node, "_node"):
            self._name_start(node._node.name)
            self._popup_start()
            self._types(node._node)
            self._scopes(node._node)
            self._popup_end()
            self._name_end()
        else:
            raise ValueError, node
            self._name(node.name)

    def visitNot(self, node):
        self.stream.write("<span class='not'>\n")
        self._keyword("not")
        self.dispatch(node.expr)
        self.stream.write("</span>")

    def visitOr(self, node):
        self.stream.write("<span class='or'>\n")
        first = 1
        for n in node.nodes:
            if not first:
                self._keyword("or")
            self.dispatch(n)
            first = 0
        self.stream.write("</span>")

    def visitPower(self, node):
        self._visitBinary(node, "power", "**")

    def visitSlice(self, node):
        self.stream.write("<span class='slice'>\n")
        self.dispatch(node.expr)
        self.stream.write("[")
        if node.lower:
            self.dispatch(node.lower)
        self.stream.write(":")
        if node.upper:
            self.dispatch(node.upper)
        # NOTE: Step?
        self.stream.write("]")
        self.stream.write("</span>\n")

    def visitSub(self, node):
        self._visitBinary(node, "sub", "-")

    def visitSubscript(self, node):
        self.stream.write("<span class='subscript'>\n")
        self.dispatch(node.expr)
        self.stream.write("[")
        first = 1
        for sub in node.subs:
            if not first:
                self.stream.write(", ")
            self.dispatch(sub)
            first = 0
        self.stream.write("]")
        self.stream.write("</span>\n")

    visitTuple = visitAssTuple

    def visitUnaryAdd(self, node):
        self._visitUnary(node, "add", "+")

    def visitUnarySub(self, node):
        self._visitUnary(node, "sub", "-")

    # Output preparation methods.

    def _text(self, text):
        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    def _attr(self, attr):
        return self._text(attr).replace("'", "&apos;").replace('"', "&quot;")

    def _url(self, url):
        return self._attr(url).replace("#", "%23").replace("-", "%2d")

    def _comment(self, comment):
        self.stream.write("<span class='comment'># %s</span>\n" % comment)

    def _keyword(self, kw):
        self.stream.write("<span class='keyword'>%s</span> " % kw)

    def _doc(self, node):
        if node.doc is not None:
            self.stream.write("<pre class='doc'>\n")
            self.stream.write('"""')
            output = textwrap.dedent(node.doc.replace('"""', '\\"\\"\\"'))
            self.stream.write(self._text(output))
            self.stream.write('"""')
            self.stream.write("</pre>\n")

    def _sequence(self, node):
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(",\n")
            self.dispatch(n)
            first = 0

    def _parameters(self, subprogram):
        first = 1
        for param, default in subprogram.params:
            if not first:
                self.stream.write(",\n")
            self._parameter(subprogram, param, default)
            first = 0
        if subprogram.star is not None:
            if not first:
                self.stream.write(", *\n")
            param, default = subprogram.star
            self._parameter(subprogram, param, default)
            first = 0
        if subprogram.dstar is not None:
            if not first:
                self.stream.write(", **\n")
            param, default = subprogram.dstar
            self._parameter(subprogram, param, default)
            first = 0

    def _parameter(self, subprogram, param, default):
        self._name_start(param)
        if hasattr(subprogram, "paramtypes"):
            self._popup_start()
            self._types_list(subprogram.paramtypes[param])
            self._popup_end()
        self._name_end()
        if default is not None and default.original is not None:
            self.stream.write("=\n")
            self.dispatch(default.original)

    def _name(self, name):
        self.stream.write("<span class='name'>%s</span>\n" % name)

    def _name_start(self, name):
        self.stream.write("<span class='name'>%s\n" % name)

    def _name_end(self):
        self.stream.write("</span>\n")

    def _popup_start(self):
        self.stream.write("<span class='popup'>\n")

    def _popup_end(self):
        self.stream.write("</span>\n")

    def _op(self, op_name, op):
        if op is not None:
            if isinstance(op, Not):
                self._invocations(op.expr)
            else:
                self._invocations(op)

    def _invocations(self, node):
        self.stream.write("<div class='invocations'>\n")
        self._invocations_list(node)
        self.stream.write("</div>\n")

    def _invocations_list(self, node):
        if hasattr(node, "invocations"):
            for invocation in node.invocations:
                fn = invocation.full_name()
                module = invocation.module.name
                name = invocation.name
                structures = [x.name for x in invocation.structures]
                self.stream.write("<div class='invocation'>")
                self.stream.write("<a href='%s.html#%s'>" % (self._url(module), self._url(fn)))
                self.stream.write(self._text(".".join([module] + structures + [name])))
                self.stream.write("</a>")
                self.stream.write("</div>\n")

    def _types(self, node):
        if hasattr(node, "types"):
            if node.types:
                self._types_list(node.types)
            else:
                self.stream.write("<div class='types'>\n")
                self.stream.write("no types\n")
                self.stream.write("</div>\n")
        else:
            self.stream.write("<div class='types'>\n")
            self.stream.write("unvisited\n")
            self.stream.write("</div>\n")

    def _types_list(self, types, style_class="types"):
        self.stream.write("<div class='%s'>\n" % style_class)
        for type in types:
            fn = type.type.full_name()
            self.stream.write("<div class='type'>")
            self.stream.write(self._text(fn)) 
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def _raises(self, node):
        if hasattr(node, "namespace") and hasattr(node.namespace, "raises") and node.namespace.raises:
            self._types_list(node.namespace.raises, style_class="raises")

    def _scopes(self, node):
        if not isinstance(node, LoadName):
            if hasattr(node, "writes") and node.writes or hasattr(node, "accesses") and node.accesses:
                self.stream.write("<div class='scopes'>\n")
                for ref in getattr(node, "writes", getattr(node, "accesses", {})).keys():
                    fn = ref.full_name()
                    self.stream.write("<div class='scope'>")
                    self.stream.write(self._text(fn)) 
                    self.stream.write("</div>\n")
                self.stream.write("</div>\n")
            if hasattr(node, "non_accesses") and node.non_accesses:
                self._types_list(node.non_accesses, style_class="non-accesses")
            if hasattr(node, "non_writes") and node.non_writes:
                self._types_list(node.non_writes, style_class="non-writes")

# Utility functions.

def flatten(lists):
    result = []
    for l in lists:
        for attr in l:
            if attr not in result:
                result.append(attr)
    return result

# Convenience functions.

def view(module, stream=None):
    viewer = Viewer(stream or sys.stdout)
    viewer.process(module.original)

def report(exc):
    viewer = Viewer(stream or sys.stdout)
    viewer.report(exc)

def browse(module, stream=None):
    browser = Browser(stream or sys.stdout)
    browser.process(module.original)

def makedoc(module, filename):
    stream = open(filename, "wb")
    try:
        browser = Browser(stream)
        browser.process(module.original)
    finally:
        stream.close()

def makedocs(module, builtins):
    try:
        dirname = "%s-docs" % module.name
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        makedoc(module, os.path.join(dirname, "%s%shtml" % (module.name, os.path.extsep)))
        makedoc(builtins, os.path.join(dirname, "%s%shtml" % (builtins.name, os.path.extsep)))
    except ViewerError, exc:
        raise

# vim: tabstop=4 expandtab shiftwidth=4
